@@ -31,7 +31,8 @@ class RequestStatus:
     async def wait_for_connection(
         self, timeout: Optional[float] = None
     ) -> AsyncConnectionInterface:
-        await self._connection_acquired.wait(timeout=timeout)
+        if self.connection is None:
+            await self._connection_acquired.wait(timeout=timeout)
         assert self.connection is not None
         return self.connection
 
@@ -208,15 +209,15 @@ class AsyncConnectionPool(AsyncRequestInterface):
             raise UnsupportedProtocol(
                 "Request URL is missing an 'http://' or 'https://' protocol."
             )
-        if scheme not in ("http", "https"):
+        if scheme not in ("http", "https", "ws", "wss"):
             raise UnsupportedProtocol(
                 f"Request URL has an unsupported protocol '{scheme}://'."
             )
 
         status = RequestStatus(request)
-        self._requests.append(status)
 
         async with self._pool_lock:
+            self._requests.append(status)
             await self._close_expired_connections()
             await self._attempt_to_acquire_connection(status)
 
@@ -229,8 +230,11 @@ class AsyncConnectionPool(AsyncRequestInterface):
                 # If we timeout here, or if the task is cancelled, then make
                 # sure to remove the request from the queue before bubbling
                 # up the exception.
-                self._requests.remove(status)
-                raise exc
+                async with self._pool_lock:
+                    # Ensure only remove when task exists.
+                    if status in self._requests:
+                        self._requests.remove(status)
+                    raise exc
 
             try:
                 response = await connection.handle_async_request(request)
@@ -273,11 +277,10 @@ class AsyncConnectionPool(AsyncRequestInterface):
         assert status.connection is not None
         connection = status.connection
 
-        if status in self._requests:
-            self._requests.remove(status)
-
         async with self._pool_lock:
             # Update the state of the connection pool.
+            if status in self._requests:
+                self._requests.remove(status)
 
             if connection.is_closed() and connection in self._pool:
                 self._pool.remove(connection)
@@ -301,19 +304,11 @@ class AsyncConnectionPool(AsyncRequestInterface):
         Close any connections in the pool.
         """
         async with self._pool_lock:
-            requests_still_in_flight = len(self._requests)
-
             for connection in self._pool:
                 await connection.aclose()
             self._pool = []
             self._requests = []
 
-            if requests_still_in_flight:
-                raise RuntimeError(
-                    f"The connection pool was closed while {requests_still_in_flight} "
-                    f"HTTP requests/responses were still in-flight."
-                )
-
     async def __aenter__(self) -> "AsyncConnectionPool":
         return self
 
@@ -349,6 +344,6 @@ class ConnectionPoolByteStream:
     async def aclose(self) -> None:
         try:
             if hasattr(self._stream, "aclose"):
-                await self._stream.aclose()  # type: ignore
+                await self._stream.aclose()
         finally:
             await self._pool.response_closed(self._status)
